from datetime import date
from typing import Dict, List, Optional, Union

from asgiref.sync import sync_to_async
from django.apps import apps
from django.db import connections
from forestadmin.datasource_django.collection import DjangoCollection
from forestadmin.datasource_django.exception import DjangoDatasourceException
from forestadmin.datasource_django.interface import BaseDjangoDatasource
from forestadmin.datasource_toolkit.exceptions import NativeQueryException
from forestadmin.datasource_toolkit.interfaces.records import RecordsDataAlias


class DjangoDatasource(BaseDjangoDatasource):
    def __init__(
        self,
        support_polymorphic_relations: bool = False,
        live_query_connection: Optional[Union[str, Dict[str, str]]] = None,
    ) -> None:
        """ Create a django datasource.
        More information here:
        https://docs.forestadmin.com/developer-guide-agents-python/data-sources/provided-data-sources/django


        Args:
            support_polymorphic_relations (bool, optional, default to `False`): Enable introspection over \
                polymorphic relation (AKA GenericForeignKey). Defaults to False.
            live_query_connection (Union[str, Dict[str, str]], optional, default to `None`): Set a connectionName to \
                use live queries. If a string is given, this connection will be map to django 'default' database. \
                Otherwise, you must use a dict `{'connectionName': 'DjangoDatabaseName'}`. \
                None doesn't enable this feature.
        """
        self._django_live_query_connections: Dict[str, str] = self._handle_live_query_connections_param(
            live_query_connection
        )
        super().__init__([*self._django_live_query_connections.keys()])

        self.support_polymorphic_relations = support_polymorphic_relations
        self._create_collections()

    def _handle_live_query_connections_param(
        self, live_query_connections: Optional[Union[str, Dict[str, str]]]
    ) -> Dict[str, str]:
        if live_query_connections is None:
            return {}

        if isinstance(live_query_connections, str):
            ret = {live_query_connections: "default"}
        else:
            ret = live_query_connections

        for forest_name, db_name in ret.items():
            if db_name not in connections:
                raise DjangoDatasourceException(
                    f"Connection to database '{db_name}' for alias '{forest_name}' is not found in django databases. "
                    f"Existing connections are {','.join([*connections])}"
                )
        return ret

    def _create_collections(self):
        models = apps.get_models(include_auto_created=True)
        for model in models:
            if model._meta.proxy is False:
                collection = DjangoCollection(self, model, self.support_polymorphic_relations)
                self.add_collection(collection)

    async def execute_native_query(
        self, connection_name: str, native_query: str, parameters: Dict[str, str]
    ) -> List[RecordsDataAlias]:
        if connection_name not in self._django_live_query_connections.keys():
            # This one should never occur while datasource composite works fine
            raise NativeQueryException(f"Native query connection '{connection_name}' is not known by DjangoDatasource.")

        def _execute_native_query():
            cursor = connections[self._django_live_query_connections[connection_name]].cursor()  # type: ignore
            try:
                # replace '\%' by '%%'
                # %(var)s is already the correct  syntax
                rows = cursor.execute(native_query.replace("\\%", "%%"), parameters)

                ret = []
                for row in rows:
                    return_row = {}
                    for i, field_name in enumerate(rows.description):
                        value = row[i]
                        if isinstance(value, date):
                            value = value.isoformat()
                        return_row[field_name[0]] = value
                    ret.append(return_row)
                return ret
            except Exception as e:
                raise NativeQueryException(str(e))

        return await sync_to_async(_execute_native_query)()



from typing import List, Optional

from asgiref.sync import sync_to_async
from django.conf import settings
from django.db import connection, connections
from django.db.models import Model
from forestadmin.agent_toolkit.forest_logger import ForestLogger
from forestadmin.agent_toolkit.utils.context import User
from forestadmin.datasource_django.interface import BaseDjangoCollection
from forestadmin.datasource_django.utils.model_introspection import DjangoCollectionFactory
from forestadmin.datasource_django.utils.native_driver_wrapper import NativeDriverWrapper, get_db_for_native_driver
from forestadmin.datasource_django.utils.polymorphic_util import DjangoPolymorphismUtil
from forestadmin.datasource_django.utils.query_factory import DjangoQueryBuilder
from forestadmin.datasource_django.utils.record_serializer import instance_to_record_data
from forestadmin.datasource_toolkit.datasources import Datasource
from forestadmin.datasource_toolkit.interfaces.fields import is_column
from forestadmin.datasource_toolkit.interfaces.query.aggregation import AggregateResult, Aggregation
from forestadmin.datasource_toolkit.interfaces.query.filter.paginated import PaginatedFilter
from forestadmin.datasource_toolkit.interfaces.query.filter.unpaginated import Filter
from forestadmin.datasource_toolkit.interfaces.query.projections import Projection
from forestadmin.datasource_toolkit.interfaces.records import RecordsDataAlias


class DjangoCollection(BaseDjangoCollection):
    def __init__(self, datasource: Datasource, model: Model, support_polymorphic_relations: bool):
        super().__init__(model._meta.db_table, datasource)
        self._model = model
        self.support_polymorphic_relations = support_polymorphic_relations
        schema = DjangoCollectionFactory.build(model, support_polymorphic_relations)
        self.add_fields(schema["fields"])
        self.enable_count()

    @property
    def model(self) -> Model:
        return self._model

    async def list(self, caller: User, filter_: PaginatedFilter, projection: Projection) -> List[RecordsDataAlias]:
        def _list():
            if self.support_polymorphic_relations:
                DjangoPolymorphismUtil.request_content_type()
            ret = [
                instance_to_record_data(item, projection, self)
                for item in DjangoQueryBuilder.mk_list(self, filter_, projection)
            ]
            if getattr(settings, "DEBUG"):
                ForestLogger.log("debug", f"SQL queries for list({len(connection.queries)}):{str(connection.queries)}")
            return ret

        return await sync_to_async(_list)()

    async def aggregate(
        self, caller: User, filter_: Optional[Filter], aggregation: Aggregation, limit: Optional[int] = None
    ) -> List[AggregateResult]:
        def _aggregate():
            if self.support_polymorphic_relations:
                DjangoPolymorphismUtil.request_content_type()
            ret = DjangoQueryBuilder.mk_aggregate(self, filter_, aggregation, limit)
            if getattr(settings, "DEBUG"):
                ForestLogger.log(
                    "debug", f"SQL queries for aggregate({len(connection.queries)}):{str(connection.queries)}"
                )
            return ret

        return await sync_to_async(_aggregate)()

    async def create(self, caller: User, data: List[RecordsDataAlias]) -> List[RecordsDataAlias]:
        projection = Projection(*[k for k in self.schema["fields"].keys() if is_column(self.schema["fields"][k])])

        def _create():
            if self.support_polymorphic_relations:
                DjangoPolymorphismUtil.request_content_type()
            ret = [instance_to_record_data(item, projection, self) for item in DjangoQueryBuilder.mk_create(self, data)]

            if getattr(settings, "DEBUG"):
                ForestLogger.log(
                    "debug", f"SQL queries for create({len(connection.queries)}):{str(connection.queries)}"
                )
            return ret

        return await sync_to_async(_create)()

    async def update(self, caller: User, filter_: Optional[Filter], patch: RecordsDataAlias) -> None:
        def _update():
            if self.support_polymorphic_relations:
                DjangoPolymorphismUtil.request_content_type()
            DjangoQueryBuilder.mk_update(self, filter_, patch)
            if getattr(settings, "DEBUG"):
                ForestLogger.log(
                    "debug", f"SQL queries for update({len(connection.queries)}):{str(connection.queries)}"
                )

        await sync_to_async(_update)()

    async def delete(self, caller: User, filter_: Optional[Filter]) -> None:
        def _delete():
            if self.support_polymorphic_relations:
                DjangoPolymorphismUtil.request_content_type()
            DjangoQueryBuilder.mk_delete(self, filter_)
            if getattr(settings, "DEBUG"):
                ForestLogger.log(
                    "debug", f"SQL queries for delete({len(connection.queries)}):{str(connection.queries)}"
                )

        await sync_to_async(_delete)()

    def get_native_driver(self) -> NativeDriverWrapper:
        db_name = get_db_for_native_driver(self.model)
        return NativeDriverWrapper(connections[db_name])



from typing import Dict, List, Optional, cast

from forestadmin.datasource_django.interface import BaseDjangoCollection
from forestadmin.datasource_toolkit.interfaces.fields import (
    PolymorphicManyToOne,
    is_polymorphic_many_to_one,
    is_reverse_polymorphic_relation,
)
from forestadmin.datasource_toolkit.interfaces.query.condition_tree.nodes.base import ConditionTree
from forestadmin.datasource_toolkit.interfaces.query.condition_tree.nodes.branch import ConditionTreeBranch
from forestadmin.datasource_toolkit.interfaces.query.condition_tree.nodes.leaf import ConditionTreeLeaf
from forestadmin.datasource_toolkit.interfaces.query.projections import Projection
from forestadmin.datasource_toolkit.interfaces.records import RecordsDataAlias


class DjangoPolymorphismProjectionUtil:
    @classmethod
    def get_polymorphism_relations(cls, projection: Projection, collection: BaseDjangoCollection) -> List[str]:
        ret = []
        for field_name, field_schema in collection.schema["fields"].items():
            if (f"{field_name}:*" in projection or field_name in projection) and is_polymorphic_many_to_one(
                field_schema
            ):
                ret.append(field_name)
        return ret

    @classmethod
    def is_polymorphism_implied(cls, projection: Projection, collection: BaseDjangoCollection) -> bool:
        fields_schema = collection.schema["fields"]
        for field, subfields in projection.relations.items():
            if subfields == ["*"] and is_polymorphic_many_to_one(fields_schema[field]):
                return True
            elif "*" not in subfields and is_reverse_polymorphic_relation(fields_schema[field]):
                return True
        return False

    @classmethod
    def is_type_field_of_generic_fk(cls, field_name: str, collection: BaseDjangoCollection) -> bool:
        generic_fk_fields = cls.get_polymorphism_relations(Projection(*collection.schema["fields"].keys()), collection)
        for poly_field in generic_fk_fields:
            schema = cast(PolymorphicManyToOne, collection.schema["fields"][poly_field])
            if field_name == schema["foreign_key_type_field"]:
                return True
        return False


class DjangoPolymorphismPatchUtil:
    @classmethod
    def replace_content_type_in_patch(
        cls, patch: RecordsDataAlias, collection: BaseDjangoCollection
    ) -> RecordsDataAlias:
        for field_name, field_schema in collection.schema["fields"].items():
            if is_polymorphic_many_to_one(field_schema) and field_schema["foreign_key_type_field"] in patch.keys():
                content_type_field = field_schema["foreign_key_type_field"]
                patch[content_type_field] = DjangoPolymorphismUtil._CONTENT_TYPE_CACHE.get(patch[content_type_field])
        return patch


class DjangoPolymorphismConditionTreeUtil:
    @classmethod
    def replace_content_type_in_condition_tree(
        cls, condition_tree: Optional[ConditionTree], collection: BaseDjangoCollection
    ) -> Optional[ConditionTree]:
        if condition_tree is None:
            return condition_tree

        if isinstance(condition_tree, ConditionTreeBranch):
            return ConditionTreeBranch(
                condition_tree.aggregator,
                [
                    cls.replace_content_type_in_condition_tree(ct, collection) for ct in condition_tree.conditions
                ],  # type:ignore
            )

        condition_tree = cast(ConditionTreeLeaf, condition_tree)
        if DjangoPolymorphismProjectionUtil.is_type_field_of_generic_fk(condition_tree.field, collection):
            return ConditionTreeLeaf(
                condition_tree.field,
                condition_tree.operator,
                DjangoPolymorphismUtil._CONTENT_TYPE_CACHE.get(condition_tree.value),  # type:ignore
            )
        return condition_tree


class DjangoPolymorphismUtil(
    DjangoPolymorphismProjectionUtil, DjangoPolymorphismPatchUtil, DjangoPolymorphismConditionTreeUtil
):
    _CONTENT_TYPE_CACHE: Optional[Dict[str, "ContentType"]] = None  # noqa:F821 # type:ignore

    @classmethod
    def request_content_type(cls):
        if cls._CONTENT_TYPE_CACHE is not None:
            return
        from django.contrib.contenttypes.models import ContentType

        cls._CONTENT_TYPE_CACHE = {}
        qs = ContentType.objects.all()
        for ct in qs:  # type: ignore
            model = ct.model_class()
            if model is not None:
                cls._CONTENT_TYPE_CACHE[model._meta.db_table] = ct

    @classmethod
    def get_collection_name_from_content_type(cls, content_type) -> str:
        if content_type is None:
            return None  # type:ignore

        for collection_name, ct in cls._CONTENT_TYPE_CACHE.items():
            if ct.id == content_type.id:
                return collection_name
